"""Defines PluginLocationRemoveStatus, PluginLocationRemoveManager, DbRemoveManager, MeltanoYmlRemoveManager and InstallationRemoveManager."""

import shutil
from abc import ABC, abstractmethod
from enum import Enum

import sqlalchemy
from meltano.core.db import project_engine
from meltano.core.plugin.error import PluginNotFoundError
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.project_plugins_service import ProjectPluginsService

from .project import Project
from .settings_store import SettingValueStore


class PluginLocationRemoveStatus(Enum):
    """Possible remove statuses."""

    REMOVED = "removed"
    ERROR = "error"
    NOT_FOUND = "not found"


class PluginLocationRemoveManager(ABC):
    """Handle removal of a plugin from a given location."""

    def __init__(self, plugin: ProjectPlugin, location):
        """Construct a PluginLocationRemoveManager instance."""
        self.plugin = plugin
        self.plugin_descriptor = f"{plugin.type.descriptor} '{plugin.name}'"
        self.location = location
        self.remove_status = None
        self.remove_message = None

    @abstractmethod
    def remove(self):
        """Abstract remove method."""
        pass

    @property
    def plugin_removed(self) -> bool:
        """Wether or not the plugin was successfully removed."""
        return self.remove_status is PluginLocationRemoveStatus.REMOVED

    @property
    def plugin_not_found(self) -> bool:
        """Wether or not the plugin was not found to remove."""
        return self.remove_status is PluginLocationRemoveStatus.NOT_FOUND

    @property
    def plugin_error(self) -> bool:
        """Wether or not an error was encountered the plugin removal process."""
        return self.remove_status is PluginLocationRemoveStatus.ERROR


class DbRemoveManager(PluginLocationRemoveManager):
    """Handle removal of a plugin's settings from the system database `plugin_settings` table."""

    def __init__(self, plugin, project):
        """Construct a DbRemoveManager instance."""
        super().__init__(plugin, "system database")
        self.plugins_settings_service = PluginSettingsService(project, plugin)
        self.session = project_engine(project)[1]

    def remove(self):
        """Remove the plugin's settings from the system database `plugin_settings` table."""
        session = self.session()
        try:
            self.plugins_settings_service.reset(
                store=SettingValueStore.DB, session=session
            )
        except sqlalchemy.exc.OperationalError as err:
            self.remove_status = PluginLocationRemoveStatus.ERROR
            self.message = err.orig
            return

        self.remove_status = PluginLocationRemoveStatus.REMOVED


class MeltanoYmlRemoveManager(PluginLocationRemoveManager):
    """Handle removal of a plugin from `meltano.yml`."""

    def __init__(self, plugin, project: Project):
        """Construct a MeltanoYmlRemoveManager instance."""
        super().__init__(plugin, str(project.meltanofile.relative_to(project.root)))
        self.project_plugins_service = ProjectPluginsService(project)

    def remove(self):
        """Remove the plugin from `meltano.yml`."""
        try:
            self.project_plugins_service.remove_from_file(self.plugin)
        except PluginNotFoundError:
            self.remove_status = PluginLocationRemoveStatus.NOT_FOUND
            return
        except OSError as err:
            self.remove_status = PluginLocationRemoveStatus.ERROR
            self.message = err.strerror
            return

        self.remove_status = PluginLocationRemoveStatus.REMOVED


class InstallationRemoveManager(PluginLocationRemoveManager):
    """Handle removal of a plugin installation from `.meltano`."""

    def __init__(self, plugin, project: Project):
        """Construct a InstallationRemoveManager instance."""
        path = project.plugin_dir(plugin, make_dirs=False)
        super().__init__(plugin, str(path.parent.relative_to(project.root)))
        self.path = path

    def remove(self):
        """Remove the plugin installation from `.meltano`."""
        if not self.path.exists():
            self.remove_status = PluginLocationRemoveStatus.NOT_FOUND
            self.message = f"{self.plugin_descriptor} not found in {self.path.parent}"
            return

        try:
            shutil.rmtree(self.path)
        except OSError as err:
            self.remove_status = PluginLocationRemoveStatus.ERROR
            self.message = err.strerror
            return

        self.remove_status = PluginLocationRemoveStatus.REMOVED



"""Base Error classes."""
import functools
import io
import logging
import subprocess
from asyncio.streams import StreamReader
from asyncio.subprocess import Process
from enum import Enum
from typing import Optional, Union


class ExitCode(int, Enum):
    OK = 0
    FAIL = 1
    NO_RETRY = 2


class Error(Exception):
    """Base exception for ELT errors."""

    def exit_code(self):
        return ExitCode.FAIL


class ExtractError(Error):
    """Error in the extraction process, like API errors."""

    def exit_code(self):
        return ExitCode.NO_RETRY


class SubprocessError(Exception):
    """Happens when subprocess exits with a resultcode != 0."""

    def __init__(
        self,
        message: str,
        process: subprocess.CompletedProcess,
        stderr: Optional[Union[str, bytes, io.TextIOBase]] = None,
    ):
        """Initialize SubprocessError."""
        self.process = process
        self._stderr = stderr or process.stderr
        super().__init__(message)

    @property
    def stderr(self) -> Optional[str]:
        """Return the output of the process to stderr."""
        if not self._stderr:
            return None
        elif isinstance(self._stderr, bytes):
            self._stderr = self._stderr.decode("utf-8")
        elif not isinstance(self._stderr, str):
            self._stderr = self._stderr.read()

        return self._stderr


class AsyncSubprocessError(Exception):
    """Happens when an async subprocess exits with a resultcode != 0."""

    def __init__(
        self,
        message: str,
        process: Process,
        stderr: Optional[str] = None,
    ):
        """Initialize AsyncSubprocessError."""
        self.process = process
        self._stderr: Union[str, StreamReader, None] = stderr or process.stderr
        super().__init__(message)

    @property
    async def stderr(self) -> Optional[str]:
        """Return the output of the process to stderr."""
        if not self._stderr:
            return None
        elif not isinstance(self._stderr, str):
            stream = await self._stderr.read()
            self._stderr = stream.decode("utf-8")

        return self._stderr


class PluginInstallError(Exception):
    """Happens when a plugin fails to install."""

    pass


class PluginInstallWarning(Exception):
    """Happens when a plugin optional optional step fails to install."""

    pass


class PluginNotInstallable(Exception):
    pass


def aggregate(error_cls):
    class Aggregate(error_cls):
        """Aggregate multiple sub-exceptions."""

        def __init__(self, exceptions: []):
            self.exceptions = exceptions

        def __str__(self):
            return "\n".join(str(e) for e in self.exceptions)

    if error_cls != Exception:
        error_cls.Aggregate = Aggregate

    return error_cls


AggregateError = aggregate(Exception)


@aggregate
class ImportError(Error):
    """
    Error in the import process, the data cannot be processed.
    """

    def __init__(self, entry):
        self.entry = entry


@aggregate
class SchemaError(Error):
    """Base exception for schema errors."""


@aggregate
class InapplicableChangeError(SchemaError):
    """Raise for inapplicable schema changes."""

    def exit_code(self):
        return ExitCode.NO_RETRY


# TODO: use as a context manager instead
class ExceptionAggregator:
    def __init__(self, etype):
        self.success = []
        self.failures = []
        self.etype = etype

    def recognize_exception(self, e) -> bool:
        return self.etype == type(e)

    def call(self, callable, *args, **kwargs):
        params = (args, kwargs)
        try:
            ret = callable(*args, **kwargs)
            self.success.append(params)
            return ret
        except Exception as e:
            if self.recognize_exception(e):
                self.failures.append((e, params))
            else:
                raise e

    def raise_aggregate(self):
        aggregate_type = AggregateError

        if hasattr(self.etype, "Aggregate"):
            aggregate_type = self.etype.Aggregate

        if len(self.failures):
            exceptions = [f[0] for f in self.failures]
            raise aggregate_type(exceptions)


def with_error_exit_code(main):
    @functools.wraps(main)
    def f(*args, **kwargs):
        try:
            main(*args, **kwargs)
        except Error as err:
            logging.error(str(err))
            exit(err.exit_code())
        except Exception as e:
            logging.error(str(e))
            raise e

    return f



"""Manage Python virtual environments."""
import asyncio
import logging
import os
import platform
import shutil
import subprocess
import sys
from asyncio.subprocess import Process
from collections import namedtuple
from pathlib import Path

from .error import AsyncSubprocessError
from .project import Project

logger = logging.getLogger(__name__)

VenvSpecs = namedtuple("VenvSpecs", ("lib_dir", "bin_dir", "site_packages_dir"))

POSIX = VenvSpecs(
    lib_dir="lib",
    bin_dir="bin",
    site_packages_dir=os.path.join("lib", f"python{sys.version[:3]}", "site-packages"),
)

NT = VenvSpecs(
    lib_dir="Lib",
    bin_dir="Scripts",
    site_packages_dir=os.path.join("Lib", "site-packages"),
)


class VirtualEnv:
    PLATFORM_SPECS = {"Linux": POSIX, "Darwin": POSIX, "Windows": NT}

    def __init__(self, root: Path):
        self.root = root
        self._specs = self.__class__.platform_specs()

    @classmethod
    def platform_specs(cls):
        system = platform.system()
        try:
            return cls.PLATFORM_SPECS[system]
        except KeyError:
            raise Exception(f"Platform {system} not supported.")

    def __getattr__(self, attr):
        spec = getattr(self._specs, attr)
        return self.root.joinpath(spec)

    def __str__(self):
        return str(self.root)


async def exec_async(*args, **kwargs) -> Process:
    """
    Run an executable asyncronously.

    :raises AsyncSubprocessError: if the command fails
    """
    run = await asyncio.create_subprocess_exec(
        *args,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        **kwargs,
    )
    await run.wait()

    if run.returncode != 0:
        raise AsyncSubprocessError("Command failed", run)

    return run


class VenvService:
    def __init__(self, project: Project, namespace: str = "", name: str = ""):
        """
        Manage isolated virtual environments.

        The methods in this class are not threadsafe.
        """
        self.project = project
        self.namespace = namespace
        self.name = name
        self.venv = VirtualEnv(self.project.venvs_dir(namespace, name))
        self.python_path = str(self.venv.bin_dir.joinpath("python"))

    async def clean_install(self, *pip_urls):
        """
        Wipe and recreate a virtual environment and install the given `pip_urls` packages in it.

        :raises: SubprocessError: if any of the commands fail.
        """
        self.clean()
        await self.create()
        await self.upgrade_pip()
        await asyncio.gather(*[self.install(pip_url) for pip_url in pip_urls])

    def clean(self):
        """Destroy the virtual environment, if it exists."""
        try:
            shutil.rmtree(str(self.venv))
            logger.debug(
                f"Removed old virtual environment for '{self.namespace}/{self.name}'"
            )
        except FileNotFoundError:
            # If the VirtualEnv has never been created before do nothing
            logger.debug("No old virtual environment to remove")
            pass

        return

    async def create(self):
        """
        Create a new virtual environment.

        :raises: SubprocessError: if the command fails.
        """
        logger.debug(f"Creating virtual environment for '{self.namespace}/{self.name}'")
        try:
            return await exec_async(
                sys.executable,
                "-m",
                "venv",
                str(self.venv),
            )
        except AsyncSubprocessError as err:
            raise AsyncSubprocessError(
                f"Could not create the virtualenv for '{self.namespace}/{self.name}'",
                err.process,
            )

    async def upgrade_pip(self):
        """
        Upgrade the `pip` package to the latest version in the virtual environment.

        :raises: SubprocessError: if the command fails.
        """
        logger.debug(f"Upgrading pip for '{self.namespace}/{self.name}'")
        try:
            return await exec_async(
                self.python_path,
                "-m",
                "pip",
                "install",
                "--upgrade",
                "pip",
                "setuptools",
                "wheel",
            )
        except AsyncSubprocessError as err:
            raise AsyncSubprocessError(
                "Failed to upgrade pip to the latest version.", err.process
            )

    async def install(self, pip_url: str):
        """
        Install a package using `pip` in the proper virtual environment.

        :raises: AsyncSubprocessError: if the command fails.
        """
        meltano_pth_path = self.venv.site_packages_dir.joinpath("meltano_venv.pth")
        if meltano_pth_path.exists():
            os.remove(meltano_pth_path)

        logger.debug(
            f"Installing '{pip_url}' into virtual environment for '{self.namespace}/{self.name}'"
        )
        try:
            return await exec_async(
                self.python_path,
                "-m",
                "pip",
                "install",
                *pip_url.split(" "),
            )
        except AsyncSubprocessError as err:
            raise AsyncSubprocessError(
                f"Failed to install plugin '{self.name}'.", err.process
            )

    def exec_path(self, executable):
        """Return the absolute path for the given binary in the virtual environment."""
        return self.venv.bin_dir.joinpath(executable)



import asyncio
import copy
import enum
import logging
import os
import subprocess
from contextlib import contextmanager
from pathlib import Path
from typing import Optional, Union

from meltano.core.logging.utils import SubprocessOutputWriter

from .error import Error, SubprocessError
from .plugin import PluginRef
from .plugin.config_service import PluginConfigService
from .plugin.project_plugin import ProjectPlugin
from .plugin.settings_service import PluginSettingsService
from .project import Project
from .project_plugins_service import ProjectPluginsService
from .venv_service import VenvService, VirtualEnv


def invoker_factory(project, plugin: ProjectPlugin, *args, **kwargs):
    cls = PluginInvoker

    if hasattr(plugin, "invoker_class"):
        cls = plugin.invoker_class

    invoker = cls(project, plugin, *args, **kwargs)

    return invoker


class InvokerError(Error):
    pass


class ExecutableNotFoundError(InvokerError):
    """Occurs when the executable could not be found"""

    def __init__(self, plugin: PluginRef, executable):
        super().__init__(
            f"Executable '{executable}' could not be found. "
            f"{plugin.type.descriptor.capitalize()} '{plugin.name}' may not have been installed yet using `meltano install {plugin.type.singular} {plugin.name}`, or the executable name may be incorrect."
        )


class InvokerNotPreparedError(InvokerError):
    """Occurs when `invoke` is called before `prepare`"""

    pass


class UnknownCommandError(InvokerError):
    """Occurs when `invoke` is called in command mode with an undefined command."""

    def __init__(self, plugin: PluginRef, command):
        """Initialize UnknownCommandError."""
        self.plugin = plugin
        self.command = command

    def __str__(self):
        """Return error message."""
        if self.plugin.supported_commands:
            supported_commands = ", ".join(self.plugin.supported_commands)
            desc = f"supports the following commands: {supported_commands}"
        else:
            desc = "does not define any commands."
        return " ".join(
            [
                f"Command '{self.command}' could not be found.",
                f"{self.plugin.type.descriptor.capitalize()} '{self.plugin.name}'",
                desc,
            ]
        )


class PluginInvoker:
    """This class handles the invocation of a `ProjectPlugin` instance."""

    class StdioSource(str, enum.Enum):  # noqa: WPS431
        """Describes the available unix style std io sources."""

        STDIN = "stdin"
        STDOUT = "stdout"
        STDERR = "stderr"

    def __init__(
        self,
        project: Project,
        plugin: ProjectPlugin,
        context: Optional[object] = None,
        output_handlers: Optional[dict] = None,
        run_dir=None,
        config_dir=None,
        venv_service: VenvService = None,
        plugins_service: ProjectPluginsService = None,
        plugin_config_service: PluginConfigService = None,
        plugin_settings_service: PluginSettingsService = None,
    ):
        self.project = project
        self.plugin = plugin
        self.context = context
        self.output_handlers = output_handlers

        self.venv_service: Optional[VenvService] = None
        if plugin.pip_url or venv_service:
            self.venv_service = venv_service or VenvService(
                project,
                name=plugin.name,
                namespace=plugin.type,
            )
        self.plugin_config_service = plugin_config_service or PluginConfigService(
            plugin,
            config_dir or self.project.plugin_dir(plugin),
            run_dir or self.project.run_dir(plugin.name),
        )

        self.plugins_service = plugins_service or ProjectPluginsService(project)
        self.settings_service = plugin_settings_service or PluginSettingsService(
            project, plugin, plugins_service=self.plugins_service
        )

        self._prepared = False
        self.plugin_config = {}
        self.plugin_config_processed = {}
        self.plugin_config_extras = {}
        self.plugin_config_env = {}

    @property
    def capabilities(self):
        # we want to make sure the capabilites are immutable from the `PluginInvoker` interface
        return frozenset(self.plugin.capabilities)

    @property
    def files(self):
        plugin_files = {**self.plugin.config_files, **self.plugin.output_files}

        return {
            _key: self.plugin_config_service.run_dir.joinpath(filename)
            for _key, filename in plugin_files.items()
        }

    def prepare(self, session):
        self.plugin_config = self.settings_service.as_dict(
            extras=False, session=session
        )
        self.plugin_config_processed = self.settings_service.as_dict(
            extras=False, process=True, session=session
        )
        self.plugin_config_extras = self.settings_service.as_dict(
            extras=True, session=session
        )
        self.plugin_config_env = self.settings_service.as_env(session=session)

        with self.plugin.trigger_hooks("configure", self, session):
            self.plugin_config_service.configure()
            self._prepared = True

    def cleanup(self):
        self.plugin_config = {}
        self.plugin_config_processed = {}
        self.plugin_config_extras = {}
        self.plugin_config_env = {}

        with self.plugin.trigger_hooks("cleanup", self):
            self._prepared = False

    @contextmanager
    def prepared(self, session):
        try:
            self.prepare(session)
            yield
        finally:
            self.cleanup()

    def exec_path(self, executable: Optional[str] = None) -> Union[str, Path]:
        """
        Return the absolute path to the executable.

        Uses the plugin executable if none is specified.
        """
        executable = executable or self.plugin.executable
        if not self.venv_service:
            if "/" not in executable.replace("\\", "/"):
                # Expect executable on path
                return executable

            # Return executable relative to project directory
            return self.project.root.joinpath(executable)

        # Return executable within venv
        return self.venv_service.exec_path(executable)

    def exec_args(self, *args, command=None, env=None):
        """Materialize the arguments to be passed to the executable.

        Raises `UnknownCommandError` if requested command is not defined.
        """
        env = env or {}
        executable = self.exec_path()
        if command:
            command_config = self.find_command(command)
            plugin_args = command_config.expanded_args(command, env)
            if command_config.executable:
                executable = self.exec_path(command_config.executable)
        else:
            plugin_args = self.plugin.exec_args(self)

        return [str(arg) for arg in (executable, *plugin_args, *args)]

    def find_command(self, name):
        """Find a Command by name. Raises `UnknownCommandError` if not defined."""
        try:
            return self.plugin.all_commands[name]
        except KeyError as err:
            raise UnknownCommandError(self.plugin, name) from err

    def env(self):
        env = {
            **self.project.dotenv_env,
            **self.settings_service.env,
            **self.plugin_config_env,
        }

        # Ensure Meltano venv is not inherited
        env.pop("VIRTUAL_ENV", None)
        env.pop("PYTHONPATH", None)
        if self.venv_service:
            # Switch to plugin-specific venv
            venv = VirtualEnv(
                self.project.venvs_dir(self.plugin.type, self.plugin.name)
            )
            venv_dir = str(venv.bin_dir)
            env["VIRTUAL_ENV"] = str(venv.root)
            env["PATH"] = os.pathsep.join([venv_dir, env["PATH"]])

        return env

    def Popen_options(self):
        return {}

    @contextmanager
    def _invoke(
        self,
        *args,
        require_preparation=True,
        env=None,
        command=None,
        **kwargs,
    ):
        env = env or {}

        if require_preparation and not self._prepared:
            raise InvokerNotPreparedError()

        with self.plugin.trigger_hooks("invoke", self, args):
            popen_options = {**self.Popen_options(), **kwargs}
            popen_env = {**self.env(), **env}
            popen_args = self.exec_args(*args, command=command, env=popen_env)
            logging.debug(f"Invoking: {popen_args}")
            logging.debug(f"Env: {popen_env}")

            try:
                yield (popen_args, popen_options, popen_env)
            except FileNotFoundError as err:
                raise ExecutableNotFoundError(
                    self.plugin, self.plugin.executable
                ) from err

    def invoke(self, *args, **kwargs):
        with self._invoke(*args, **kwargs) as (popen_args, popen_options, popen_env):
            return subprocess.Popen(popen_args, **popen_options, env=popen_env)

    async def invoke_async(self, *args, **kwargs):
        with self._invoke(*args, **kwargs) as (popen_args, popen_options, popen_env):
            return await asyncio.create_subprocess_exec(
                *popen_args,
                **popen_options,
                env=popen_env,
            )

    def dump(self, file_id):
        try:
            with self._invoke():
                return self.files[file_id].read_text()
        except ExecutableNotFoundError as err:
            # Unwrap FileNotFoundError
            raise err.__cause__

    def add_output_handler(self, src: str, handler: SubprocessOutputWriter):
        """Append an output handler for a given stdio stream.

        Args:
            src: stdio source you'd like to subscribe, likely either 'stdout' or 'stderr'
            handler: either a StreamWriter or an object matching the utils.SubprocessOutputWriter proto
        """
        if self.output_handlers:
            self.output_handlers[src].append(handler)
        else:
            self.output_handlers = {src: [handler]}



"""Create the EmbedToken table

Revision ID: ceb00d7ff3bd
Revises: 990c0665f3ce
Create Date: 2020-01-27 09:59:36.427138

"""
import sqlalchemy as sa
from alembic import op

# revision identifiers, used by Alembic.
revision = "ceb00d7ff3bd"
down_revision = "990c0665f3ce"
branch_labels = None
depends_on = None


def upgrade():
    op.create_table(
        "embed_tokens",
        sa.Column("id", sa.Integer, primary_key=True),
        sa.Column("token", sa.String(64), unique=True),
        sa.Column("resource_id", sa.String(), nullable=False),
        sa.Column("created_at", sa.DateTime),
    )


def downgrade():
    op.drop_table("embed_tokens")



import logging
import sys

import click
from meltano.core.db import project_engine
from meltano.core.error import SubprocessError
from meltano.core.plugin import PluginType
from meltano.core.plugin_invoker import UnknownCommandError, invoker_factory
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.tracking import GoogleAnalyticsTracker

from . import cli
from .params import pass_project
from .utils import CliError, propagate_stop_signals

logger = logging.getLogger(__name__)


@cli.command(
    context_settings=dict(ignore_unknown_options=True, allow_interspersed_args=False)
)
@click.option(
    "--plugin-type", type=click.Choice(PluginType.cli_arguments()), default=None
)
@click.option(
    "--dump",
    type=click.Choice(["catalog", "config"]),
    help="Dump content of generated file",
)
@click.option(
    "--list-commands",
    is_flag=True,
    help="List the commands supported by the plugin",
)
@click.argument("plugin_name", metavar="PLUGIN_NAME[:COMMAND_NAME]")
@click.argument("plugin_args", nargs=-1, type=click.UNPROCESSED)
@pass_project(migrate=True)
def invoke(project, plugin_type, dump, list_commands, plugin_name, plugin_args):
    """Invoke the plugin's executable with specified arguments."""
    try:
        plugin_name, command_name = plugin_name.split(":")
    except ValueError:
        command_name = None

    plugin_type = PluginType.from_cli_argument(plugin_type) if plugin_type else None

    _, Session = project_engine(project)
    session = Session()
    plugins_service = ProjectPluginsService(project)
    plugin = plugins_service.find_plugin(
        plugin_name, plugin_type=plugin_type, invokable=True
    )

    if list_commands:
        do_list_commands(plugin)
        return

    try:
        invoker = invoker_factory(project, plugin, plugins_service=plugins_service)
        with invoker.prepared(session):
            if dump:
                dump_file(invoker, dump)
                exit_code = 0
            else:
                handle = invoker.invoke(*plugin_args, command=command_name)
                with propagate_stop_signals(handle):
                    exit_code = handle.wait()

    except UnknownCommandError as err:
        raise click.BadArgumentUsage(err) from err
    except SubprocessError as err:
        logger.error(err.stderr)
        raise
    finally:
        session.close()

    tracker = GoogleAnalyticsTracker(project)
    tracker.track_meltano_invoke(
        plugin_name=plugin_name, plugin_args=" ".join(plugin_args)
    )

    sys.exit(exit_code)


def do_list_commands(plugin):
    """List the commands supported by plugin."""
    if not plugin.supported_commands:
        click.secho(
            f"Plugin '{plugin.name}' does not define any commands.", fg="yellow"
        )
        return

    descriptions = {
        f"{plugin.name}:{cmd}": props.description
        for cmd, props in plugin.all_commands.items()
    }
    column_len = max(len(name) for name in descriptions.keys()) + 2
    for name, desc in descriptions.items():
        click.secho(name.ljust(column_len, " "), fg="blue", nl=False)
        click.echo(desc)


def dump_file(invoker, file_id):
    try:
        content = invoker.dump(file_id)
        print(content)
    except FileNotFoundError as err:
        raise CliError(f"Could not find {file_id}") from err
    except Exception as err:
        raise CliError(f"Could not dump {file_id}: {err}") from err



import re
from copy import deepcopy
from datetime import datetime
from enum import Enum
from typing import Dict, List, Tuple

import networkx as nx
from networkx.readwrite import json_graph
from pypika import AliasedQuery, Criterion, Field, Interval, Order, Query, Table
from pypika import functions as fn


class ParseError(Exception):
    pass


class EmptyQuery(Exception):
    pass


class MeltanoFilterExpressionType(str, Enum):
    Unknown = "UNKNOWN"
    LessThan = "less_than"
    LessOrEqualThan = "less_or_equal_than"
    EqualTo = "equal_to"
    GreaterOrEqualThan = "greater_or_equal_than"
    GreaterThan = "greater_than"
    Like = "like"
    IsNull = "is_null"
    IsNotNull = "is_not_null"

    def __hash__(self):
        return hash(self.value)

    def __eq__(self, other):
        return self.value.lower() == other.value.lower()

    @classmethod
    def parse(cls, value: str):
        try:
            return next(e for e in cls if e.value.lower() == value)
        except StopIteration:
            return cls.Unknown


class MeltanoBase:
    def __init__(self, definition: Dict = {}):
        self._definition = definition

    def __getattr__(self, attr: str):
        return self._definition.get(attr, None)

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name}>"


class MeltanoDesign(MeltanoBase):
    """
    An internal representation of a design loaded from an m5oc file.

    It provides access to the Tables and Joins defined, together with additional
      metada for the design: {name, label, description, graph}
    """

    def __init__(self, definition: Dict) -> None:
        # Overriding the init method to force a definition for Designs
        super().__init__(definition)

    def tables(self) -> List[MeltanoBase]:
        tables = [
            MeltanoTable(self._definition.get("related_table", None), design=self)
        ]
        for join in self._definition.get("joins", []):
            table = MeltanoTable(join["related_table"], design=self)
            table.source_name = join["name"]
            tables.append(table)

        return tables

    def joins(self) -> List[MeltanoBase]:
        return [
            MeltanoJoin(definition, design=self)
            for definition in self._definition.get("joins", [])
        ]

    def get_table(self, name: str) -> MeltanoBase:
        return next(t for t in self.tables() if t.name == name)

    def get_join(self, name: str) -> MeltanoBase:
        return next(t for t in self.joins() if t.name == name)

    def find_table(self, name: str) -> MeltanoBase:
        try:
            return next(t for t in self.tables() if t.find_source_name() == name)
        except StopIteration:
            raise ParseError(f"Table {name} not found in design {self.name}")

    def find_attribute(self, key: str) -> MeltanoBase:
        try:
            return next(
                qa for t in self.tables() for qa in t.attributes() if qa.alias() == key
            )
        except StopIteration:
            raise ParseError(f"Attribute {key} not found in design {self.name}")

    def find_source_name(self, table_name: str) -> str:
        """
        Given the name of a table, get its source name:
        - If it is the base table for this design then return the name of the design
        - If it is part of a join then return the name of the join
        """
        if self._definition["related_table"]["name"] == table_name:
            return self.name

        try:
            return next(
                j["name"]
                for j in self._definition.get("joins", [])
                if j["related_table"]["name"] == table_name
            )
        except StopIteration:
            raise ParseError(f"Table {table_name} not found in design {self.name}")


class MeltanoTable(MeltanoBase):
    """
    An internal representation of a Table defined in a design.

    It provides access to the Columns and Aggregates defined, together with additional
      metada for the Table: {name, sql_table_name}
    """

    def __init__(self, definition: Dict = {}, design: MeltanoDesign = None) -> None:
        self.design = design

        # We are going to use class Table in 2 ways:
        # 1. As part of a design, with everything stored in self._definition
        # 2. As a stand alone class with no self._definition where we'll manually
        #     store everything in the following attributes
        self._attributes = {}
        self._columns = []
        self._aggregates = []
        self._timeframes = []

        # Optional Primary Keys stored for tables used in Queries
        # Those are the PK columns of the table that have not been directly
        #  requested in the Query
        self.optional_pkeys = []
        self.sql_table_name = definition.get("sql_table_name", None)

        # When a MeltanoTable is used as part of a MeltanoQuery,
        #  we also need to store the column definitions for not selected columns
        #  that have filters, so that we can use them when constructing the query
        self._unselected_columns_with_filters = []

        super().__init__(definition)

    def columns(self) -> List[MeltanoBase]:
        return [
            MeltanoColumn(definition=definition, table=self)
            for definition in self._definition.get("columns", [])
        ] + self._columns

    def unselected_columns_with_filters(self) -> List[MeltanoBase]:
        return self._unselected_columns_with_filters

    def primary_keys(self) -> List[MeltanoBase]:
        return (
            list(filter(lambda x: x.primary_key, self.columns())) + self.optional_pkeys
        )

    def aggregates(self) -> List[MeltanoBase]:
        return [
            MeltanoAggregate(definition=definition, table=self)
            for definition in self._definition.get("aggregates", [])
        ] + self._aggregates

    def timeframes(self) -> List[MeltanoBase]:
        return [
            MeltanoTimeframe(definition=definition, table=self)
            for definition in self._definition.get("timeframes", [])
        ] + self._timeframes

    def timeframe_periods(self) -> List[MeltanoBase]:
        return [tp for timeframe in self.timeframes() for tp in timeframe.periods()]

    def attributes(self) -> List[MeltanoBase]:
        return (
            self.columns()
            + self.aggregates()
            + self.timeframes()
            + self.timeframe_periods()
        )

    def get_column(self, name: str) -> MeltanoBase:
        return next((c for c in self.columns() if c.name == name), None)

    def get_aggregate(self, name: str) -> MeltanoBase:
        return next((a for a in self.aggregates() if a.name == name), None)

    def get_aggregate_by_column_name(self, name: str) -> MeltanoBase:
        return next((a for a in self.aggregates() if a.column_name() == name), None)

    def get_timeframe(self, name: str) -> MeltanoBase:
        return next((t for t in self.timeframes() if t.name == name), None)

    def get_attribute(self, name: str) -> MeltanoBase:
        """
        Given an attribute name, find if there is a column / aggregate / timeframe
         with that name and return it
        """
        if self.get_column(name):
            return self.get_column(name)
        elif self.get_aggregate(name):
            return self.get_aggregate(name)
        elif self.get_timeframe(name):
            return self.get_timeframe(name)
        else:
            raise ParseError(f"Attribute {name} not found in Meltano Table {self.name}")

    def add_column(self, column) -> None:
        self._columns.append(column)

    def add_column_with_filter(self, column) -> None:
        self._unselected_columns_with_filters.append(column)

    def add_aggregate(self, aggregate) -> None:
        self._aggregates.append(aggregate)

    def add_timeframe(self, timeframe) -> None:
        self._timeframes.append(timeframe)

    def find_source_name(self) -> str:
        """
        Get the source name for this table:
        (1) if a source name has been set for this table then return that name
        (2) Otherwise, if it is defined in a design, then get it from the design
        (3) Finally, default to MeltanoTable.name if everything else is missing
        """
        if self.source_name:
            return self.source_name
        elif self.design:
            return self.design.find_source_name(self.name)
        else:
            return self.name

    def copy_metadata(self, other_table) -> None:
        self.name = other_table.name
        self.sql_table_name = other_table.sql_table_name
        self.source_name = other_table.source_name

    def __getattr__(self, attr: str):
        try:
            return self._attributes[attr]
        except KeyError:
            return super().__getattr__(attr)

    def __setattr__(self, name, value):
        if name == "sql_table_name":
            if value is not None:
                try:
                    schema, name = value.split(".")
                    self._attributes["schema"] = schema
                    self._attributes["sql_table_name"] = name
                except ValueError:
                    self._attributes["schema"] = None
                    self._attributes["sql_table_name"] = value
        elif name in ["name", "schema", "source_name"]:
            self._attributes[name] = value
        else:
            super().__setattr__(name, value)


class MeltanoColumn(MeltanoBase):
    """
    An internal representation of a Column defined in a Table.

    It provides access to the metada for the Column:
      {label, name, type, primary_key, hidden, sql}
    together with helper functions for getting a qualified name for the Column
     and a proper alias that can be used in queries
    """

    def __init__(self, table: MeltanoTable, definition: Dict = {}) -> None:
        self.table = table
        self.attribute_type = "column"

        # Used for manually setting the metadata attributes {label, name, type, primary_key, hidden, sql}
        self._attributes = {}

        super().__init__(definition)

    def alias(self) -> str:
        return f"{self.table.find_source_name()}.{self.column_name()}"

    def column_name(self) -> str:
        if self.sql:
            (table, column) = self.sql.split(".")
        else:
            column = self.name

        return column

    def copy_metadata(self, other_column) -> None:
        self.label = other_column.label
        self.name = other_column.name
        self.type = other_column.type
        self.primary_key = other_column.primary_key
        self.hidden = other_column.hidden
        self.sql = other_column.sql

    def __getattr__(self, attr: str):
        try:
            return self._attributes[attr]
        except KeyError:
            return super().__getattr__(attr)

    def __setattr__(self, name, value):
        if name in ["label", "name", "type", "primary_key", "hidden", "sql"]:
            self._attributes[name] = value
        else:
            super().__setattr__(name, value)

    def __repr__(self):
        return f"<{self.__class__.__name__} name={self.name}, alias={self.alias()}>"


class MeltanoAggregate(MeltanoBase):
    """
    An internal representation of an Aggregate defined in a Table.

    It provides access to the metada for the Aggregate:
      {name, type (one of {sum, count, avg}), label, description, sql}
    together with helper functions for getting a qualified name for the Column,
     a proper alias that can be used in queries, the column name used in the aggregate
     and the qualified sql expression to be used in queries.
    """

    def __init__(self, table: MeltanoTable, definition: Dict = {}) -> None:
        self.table = table
        self.attribute_type = "aggregate"

        # Used for manually setting the metadata attributes {name, type, label, description, sql}
        self._attributes = {}

        super().__init__(definition)

    def alias(self) -> str:
        return f"{self.table.find_source_name()}.{self.name}"

    def qualified_sql(self, base_table: str = None, pika_table=None) -> fn.Coalesce:
        """
        Return the aggregate as a fully qualified SQL clause defined
          as a pypika Aggregate

        The base_table defines whether the Aggregate is evaluated over its
          own table or an intermediary table (e.g. resulting from a with clause)

        If the Aggregate is evaluated over its own table, then things are simple
          as we just want something like: COUNT("region"."id") "region.count"

        But if the attribute has been selected in a previous with clause, then
          its name is its alias as if it was a column (e.g. "region.id")
          and the attribute is selected by the base table, so we want something
          like: COUNT("base_table"."region.id") "region.count"

        On top of that, most of the times we define aggregates as: {{TABLE}}.id
          so we have to first replace {{TABLE}} with its own table name
        """
        if pika_table:
            field = Field(self.column_name(), table=pika_table)
        elif base_table and base_table != self.table.find_source_name():
            sql = re.sub(
                r"\{\{TABLE\}\}",
                self.table.find_source_name(),
                self.sql,
                flags=re.IGNORECASE,
            )
            field = Field(sql, table=Table(base_table))
        else:
            table = Table(self.table.find_source_name())
            field = Field(self.column_name(), table=table)

        if self.type == "sum":
            return fn.Coalesce(fn.Sum(field), 0, alias=self.alias())
        elif self.type == "count":
            return fn.Coalesce(fn.Count(field), 0, alias=self.alias())
        elif self.type == "avg":
            return fn.Coalesce(fn.Avg(field), 0, alias=self.alias())
        elif self.type == "max":
            return fn.Coalesce(fn.Max(field), 0, alias=self.alias())
        elif self.type == "min":
            return fn.Coalesce(fn.Min(field), 0, alias=self.alias())

    def column_name(self) -> str:
        (table, column) = self.sql.split(".")
        return column

    def column_alias(self) -> str:
        return f"{self.table.find_source_name()}.{self.column_name()}"

    def copy_metadata(self, other_aggregate) -> None:
        self.name = other_aggregate.name
        self.type = other_aggregate.type
        self.label = other_aggregate.label
        self.description = other_aggregate.description
        self.sql = other_aggregate.sql

    def __getattr__(self, attr: str):
        try:
            return self._attributes[attr]
        except KeyError:
            return super().__getattr__(attr)

    def __setattr__(self, name, value):
        if name in ["name", "type", "label", "description", "sql"]:
            self._attributes[name] = value
        else:
            super().__setattr__(name, value)


class MeltanoTimeframe(MeltanoBase):
    """
    An internal representation of a Timeframe defined in a Table.

    It provides access to the metada for the Timeframe:
      {name, periods:[], label, description, sql}
    """

    def __init__(self, table: MeltanoTable, definition: Dict = {}) -> None:
        self.table = table
        self.attribute_type = "timeframe"

        # Used for manually setting the metadata attributes
        #  {name, periods:[], label, description, sql}
        self._attributes = {}
        self._periods = definition.get("periods", [])

        super().__init__(definition)

    def periods(self) -> List[MeltanoBase]:
        return [
            MeltanoTimeframePeriod(definition=definition, timeframe=self)
            for definition in self._periods
        ]

    def column_name(self) -> str:
        (table, column) = self.sql.split(".")
        return column

    def alias(self) -> str:
        return f"{self.table.find_source_name()}.{self.column_name()}"

    def copy_metadata(self, other_timeframe, requested_periods=None) -> None:
        """
        Copy the metadata from another timeframe, only keeping the
         periods defined by requested_periods.

        Allows us to copy everything from the Design, by keeping only the
         periods requested in a Query.
        """
        self.name = other_timeframe.name

        # Careful cause requested periods could be []
        if requested_periods is not None:
            self._periods = [
                period
                for request in requested_periods
                for period in other_timeframe._periods
                if (
                    (period["label"] == request["label"])
                    if "label" in request
                    else (period["name"] == request["name"])
                )
            ]
        else:
            self._periods = other_timeframe._periods

        self.label = other_timeframe.label
        self.description = other_timeframe.description
        self.sql = other_timeframe.sql

    def __getattr__(self, attr: str):
        try:
            return self._attributes[attr]
        except KeyError:
            return super().__getattr__(attr)

    def __setattr__(self, name, value):
        if name in ["name", "label", "description", "sql"]:
            self._attributes[name] = value
        else:
            super().__setattr__(name, value)


class MeltanoTimeframePeriod(MeltanoBase):
    """
    An internal representation of a Timeframe period defined in a Table.

    It provides access to the metada for the Timeframe period:
      {name, label, part}
    """

    def __init__(self, timeframe: MeltanoTimeframe, definition: Dict = {}) -> None:
        self.timeframe = timeframe
        self.table = timeframe.table
        self.attribute_type = "timeframe_period"

        # Used for manually setting the metadata attributes
        #  {name, label, part}
        self._attributes = {}

        super().__init__(definition)

    def column_name(self) -> str:
        return f"{self.timeframe.column_name()}.{self.name}"

    def alias(self) -> str:
        return f"{self.timeframe.alias()}.{self.name}"

    def attribute_label(self) -> str:
        label = self.label or self.name

        if not label:
            raise ParseError(f"Requested period {period} has no name.")

        return f"{self.timeframe.label}: {label}"

    def qualified_sql(self, base_table: str = None, pika_table=None) -> fn.Coalesce:
        """
        Return the period as a fully qualified SQL clause defined as a pypika Extract clause

        The base_table defines whether the Extract is evaluated over its
          own table or an intermediary table (e.g. resulting from a with clause)

        Check MeltanoAggregate.qualified_sql() for more details
        """
        if pika_table:
            field = Field(self.timeframe.column_name(), table=pika_table)
        elif base_table and base_table != self.table.find_source_name():
            sql = re.sub(
                r"\{\{TABLE\}\}",
                self.table.find_source_name(),
                self.timeframe.sql,
                flags=re.IGNORECASE,
            )
            field = Field(sql, table=Table(base_table))
        else:
            table = Table(self.table.find_source_name())
            field = Field(self.timeframe.column_name(), table=table)

        return fn.Extract(self.part, field, alias=self.alias())

    def __getattr__(self, attr: str):
        try:
            return self._attributes[attr]
        except KeyError:
            return super().__getattr__(attr)

    def __setattr__(self, name, value):
        if name in ["name", "label", "part"]:
            self._attributes[name] = value
        else:
            super().__setattr__(name, value)


class MeltanoJoin(MeltanoBase):
    """
    An internal representation of a Join defined in a Design.

    It provides access to the metada for the Join:
      {name, related_table, label, sql_on, relationship}
    """

    def __init__(self, definition: Dict = {}, design: MeltanoDesign = None) -> None:
        self.design = design
        super().__init__(definition)


class MeltanoFilter(MeltanoBase):
    """
    An internal representation of a Filter (WHERE or HAVING clause)
    defined in a MeltanoQuery.

    definition: {key, "expression", "value"}
    legacy definition: {"source_name", "name", "expression", "value"}

    Where:
    + "source name" is the way we call a table in a design:
      - If it is the base table for the design then it is name of the design
      - If it is part of a join then it is the name of the join
      The reason for that is that we can have the same table multiple times in
       a query (e.g. multiple joins or a self join of the base table with its self),
       so we use the source_name in order to know which version of the table we
       refer to.
    + "name" is the name of the attribute accessed (column, aggregate or timeframe)
    + "expression" is one of the supported expressions defined in MeltanoFilterExpressionType
    + "value" is the value we are going to use in the expression (for non unary expressions)

    In the most simple case, the idea is to use MeltanoFilter in order to generate a simple clause:
      table_alias.name {expression} value
    e.g. gitlab_stats_per_user.project_name = 'Meltano'
     or  COALESCE(SUM("gitlab_stats_per_user"."total_issues_authored"),0) > 5

    Because we are using this dynamically over various intermediary tables,
     in reality we use match(...) to find the Filters that match a specific attribute
     and then criterion(...) to generate the proper clause based on a Pypika Field
     (the Pypika Field knows at run time the table, the clause is evaluated against)

    It provides access to the metada for the Filter:
      {key, expression, value}
    """

    def __init__(
        self,
        definition: Dict = {},
        design: MeltanoDesign = None,
        pivot_date: str = None,
    ) -> None:
        definition = deepcopy(definition)

        # The design is used for filters in queries against specific designs
        #  to validate that all the tables and attributes (columns/aggregates)
        #  are properly defined in the design
        self.design = design

        self.expression_type = MeltanoFilterExpressionType.parse(
            definition["expression"]
        )

        self.validate(definition)

        # Some Query definitions may have a pivot_date set in the `today` param
        # We use that pivot_date to define relative date filter
        #  arounf that date and not NOW()
        self.pivot_date = pivot_date

        # Check if there are relative date filter definitions, parse them
        #  and generate the proper SQL clauses
        self.parse_relative_date_filters(definition)

        super().__init__(definition)

    def is_relative_date_expression(self, value: str) -> bool:
        """
        Check if a value is a relative date expression: [+-]N[dmy]
        """
        return value is not None and re.match("^[+-][0-9]+[dmy]$", f"{value}")

    def validate(self, definition: Dict) -> None:
        """
        Validate the Filter definition
        """

        source_name = definition.pop("source_name", None)
        attribute_name = definition.pop("name", None)
        if self.design and source_name is not None and attribute_name is not None:
            table_def = self.design.find_table(source_name)
            query_attribute = table_def.get_attribute(attribute_name)
            definition["key"] = query_attribute.alias()

        key = definition.get("key", None)
        if key is None:
            raise ParseError(
                f"An attribute key was not provided for filter '{definition}'."
            )

        if self.expression_type == MeltanoFilterExpressionType.Unknown:
            raise NotImplementedError(
                f"Unknown filter expression: {definition['expression']}."
            )

        if (
            definition.get("value", None) is None
            and self.expression_type != MeltanoFilterExpressionType.IsNull
            and self.expression_type != MeltanoFilterExpressionType.IsNotNull
        ):
            raise ParseError(
                f"Filter expression: {self.expression_type} needs a non-empty value."
            )

        if (
            self.is_relative_date_expression(definition.get("value", None))
            and not self.design
        ):
            raise ParseError(
                f"[+-]N[dmy] filter expressions require a design to be set for MeltanoFilter"
            )

        if self.design:
            # Will raise ParseError if not found
            attribute_def = self.design.find_attribute(key)

            if self.is_relative_date_expression(definition.get("value", None)):
                if attribute_def.attribute_type == "aggregate":
                    raise ParseError(
                        f"You can't use '[+-]N[dmy]' filter expressions with aggregate attributes"
                    )
                elif attribute_def.type not in ["date", "time"]:  # noqa: WPS220, WPS510
                    raise ParseError(
                        f"[+-]N[dmy] filter expressions require a column atribute of type date or time"
                    )

    def parse_relative_date_filters(self, definition: Dict) -> None:
        """
        Check if there are relative date filter definitions: [+-]N[dmy]

        If so, parse them and generate the proper SQL clauses
        """
        if not self.is_relative_date_expression(definition.get("value", None)):
            # We only care about filter definitions of the type: [+-]N[dmy]
            return

        attribute_def = self.design.find_attribute(definition.get("key", None))

        old_value = definition.get("value", None)

        # Start from the Pivot Date: It can be a preset date or NOW()
        if self.pivot_date:
            pivot_date = fn.Date(self.pivot_date)
        else:
            pivot_date = fn.Date(fn.Now())

        # Then add/substract the Interval period (N) provided in [+-]N[dmy]
        interval_period = int(old_value[1:-1])

        if interval_period == 0:
            # If we have no interval period, we can really simplify the expression
            if attribute_def.type == "date":
                new_value = pivot_date
            elif attribute_def.type == "time":
                if self.expression_type == MeltanoFilterExpressionType.LessOrEqualThan:
                    new_value = (
                        pivot_date
                        + Interval(hours=23)
                        + Interval(minutes=59)
                        + Interval(seconds=59)
                        + Interval(microseconds=999_999)
                    )
                else:
                    new_value = pivot_date
        else:
            # If there is an interval set, use the [dmy] to generate the interval
            if "d" in old_value:
                interval = Interval(days=interval_period)
            elif "m" in old_value:
                interval = Interval(months=interval_period)
            elif "y" in old_value:
                interval = Interval(years=interval_period)

            # End then add or substract accordingly
            if attribute_def.type == "date":
                # Dates are simple as they are always set as they are
                # We just have to convert the timestamp expression to a Date
                if "+" in old_value:
                    new_value = fn.Date(pivot_date + interval)
                else:
                    new_value = fn.Date(pivot_date - interval)
            elif attribute_def.type == "time":
                # Timestamps are a little bit more involved in case of <= expressions
                if "+" in old_value:
                    new_value = pivot_date + interval
                else:
                    new_value = pivot_date - interval

                if self.expression_type == MeltanoFilterExpressionType.LessOrEqualThan:
                    new_value = (
                        new_value
                        + Interval(hours=23)
                        + Interval(minutes=59)
                        + Interval(seconds=59)
                        + Interval(microseconds=999_999)
                    )

        definition["value"] = new_value

    def match(self, attribute: MeltanoBase) -> bool:
        """
        Return True if this filter is defined for the given attribute
        """
        return self.key == attribute.alias()

    def criterion(self, field: Field) -> Criterion:
        """
        Generate the Pypika Criterion for this filter

        field: Pypika Field as we don't know the base table this filter is
                evaluated against (is it the base table or an intermediary table?)

        We have to use the following cases as PyPika does not allow an str
         representation of the clause on its where() and having() functions
        """
        criterion_strategies = {
            MeltanoFilterExpressionType.LessThan: lambda f: f < self.value,
            MeltanoFilterExpressionType.LessOrEqualThan: lambda f: f <= self.value,
            MeltanoFilterExpressionType.EqualTo: lambda f: f == self.value,
            MeltanoFilterExpressionType.GreaterOrEqualThan: lambda f: f >= self.value,
            MeltanoFilterExpressionType.GreaterThan: lambda f: f > self.value,
            MeltanoFilterExpressionType.Like: lambda f: f.like(self.value),
            MeltanoFilterExpressionType.IsNull: lambda f: f.isnull(),
            MeltanoFilterExpressionType.IsNotNull: lambda f: f.notnull(),
        }

        try:
            return criterion_strategies[self.expression_type](field)
        except KeyError:
            raise NotImplementedError(
                f"Unknown filter expression_type: {self.expression_type}."
            )


class MeltanoQuery(MeltanoBase):
    """
    An internal representation of a Query Definition sent to meltano for evaluation.

    The definition dictionary is parsed and the information is stored in a
    structured way:
     - Tables with only the Columns, Timeframes and Aggregates requested
     - Filters
     - The join order

    It provides also access to additional metada for the query:
      {run (whether to run the query or not), order, limit}
    and support functions for fetching all the columns and aggregates requested,
    the attributes used in the group by, etc
    """

    def __init__(self, definition: Dict, design_helper, schema: str = None) -> None:
        self.design_helper = design_helper

        # The Meltano Design this Query has been built for
        self.design = MeltanoDesign(design_helper.design)

        # The schema used at run time.
        # It is used in order to dynamically lookup the tables at that schema
        #  if it is not None
        self.schema = schema

        # A collection of all tables that are defined in the Query
        #  with the Columns, Timeframes and Aggregates defined in the Query
        self.tables = []

        # A collection of all the column and aggregate filters in the query
        # column_filters --> WHERE clauses that use columns selected in the query
        # aggregate_filters --> HAVING clauses that use aggregates computed in the query
        self.column_filters = []
        self.aggregate_filters = []

        # The proper join order that will be used for the query (list of table names)
        self.join_order = []

        self.validate_definition(definition)
        self.parse_definition(definition)

        super().__init__(definition)

    def validate_definition(self, definition: Dict) -> None:
        """
        Validate that the definition is properly formatted
        """

        self.validate_table_definition(definition)

        joins = definition.get("joins")
        if joins and not isinstance(joins, list):
            raise ParseError(f"Query definition property `joins` must be a list")

        for join_definition in joins:
            self.validate_table_definition(join_definition)

        order = definition.get("order")
        if order and not isinstance(order, list):
            raise ParseError(f"Query definition property `order` must be a list")

        filters = definition.get("filters", None)
        if filters:
            columns = filters.get("columns")
            if columns and not isinstance(columns, list):
                raise ParseError(
                    f"Query definition property `filters[columns]` must be a list"
                )

            aggregates = filters.get("aggregates")
            if aggregates and not isinstance(aggregates, list):
                raise ParseError(
                    f"Query definition property `filters[aggregates]` must be a list"
                )

    def validate_table_definition(self, definition: Dict) -> None:
        """
        Validate that the query definition for a specific table is properly formatted
        """
        if not isinstance(definition.get("name"), str):
            raise ParseError(f"Query definition property `name` must be a string")

        if not isinstance(definition.get("columns"), list):
            raise ParseError(f"Query definition property `columns` must be a list")

        if not isinstance(definition.get("aggregates"), list):
            raise ParseError(f"Query definition property `aggregates` must be a list")

        timeframes = definition.get("timeframes")
        if timeframes and not isinstance(timeframes, list):
            raise ParseError(f"Query definition property `timeframes` must be a list")

        if definition.get("today", None):
            try:
                datetime.strptime(definition.get("today", None), "%Y-%m-%d")
            except ValueError:
                raise ValueError("Incorrect pivot_date format, should be YYYY-MM-DD")

    def parse_definition(self, definition: Dict) -> None:
        """
        Parse the Query definition and generate the Tables, join_order, etc for
         easy access and lookups.
        """

        # Parse and store the pivot date provided in the Query definition
        pivot_date = definition.get("today", None)

        # Parse and store the provided filters
        filters = definition.get("filters", None)

        if filters:
            for column_filter in filters.get("columns", []):
                # Generate (and automaticaly validate) the filter and then store it
                mf = MeltanoFilter(
                    definition=column_filter, design=self.design, pivot_date=pivot_date
                )
                self.column_filters.append(mf)

            for aggregate_filter in filters.get("aggregates", []):
                # Generate (and automaticaly validate) the filter and then store it
                mf = MeltanoFilter(
                    definition=aggregate_filter,
                    design=self.design,
                    pivot_date=pivot_date,
                )
                self.aggregate_filters.append(mf)

        # Find the tables defined in the Query and add them as Table Objects
        #  with the proper Columns, Aggregates and Timeframe Objects
        primary_definition = {
            "name": definition.get("name", None),
            "columns": definition.get("columns", None),
            "aggregates": definition.get("aggregates", None),
            "timeframes": definition.get("timeframes", None),
        }
        related_definitions = [primary_definition] + definition.get("joins", [])

        # get the graph defined in the design: it is used for finding the
        #  shortest path and adding missing tables in case a required table
        #  is missing
        design_graph = json_graph.node_link_graph(self.design.graph)

        for related_def in related_definitions:
            # Find the requested table in the Design
            table_def = self.design.find_table(related_def["name"])

            if table_def:
                table = MeltanoTable(design=self.design)
                table.copy_metadata(table_def)

                # Add the columns
                for column in related_def.get("columns", []):
                    column_def = table_def.get_column(column)
                    if column_def:
                        c = MeltanoColumn(table=table)
                        c.copy_metadata(column_def)
                        table.add_column(c)
                    else:
                        raise ParseError(
                            f"Requested column {table.name}.{column} is not defined in the design"
                        )

                # Also add all the columns that are referenced in column filters
                #  for this table that are not also selected.
                # This is important in order to:
                # (1) Know that a join is required for a table with no selected columns
                # (2) Make available the missing column_def in get_query()
                for column_filter in self.column_filters:
                    # Will raise ParseError if not found
                    attribute_def = self.design.find_attribute(column_filter.key)
                    if (
                        attribute_def.attribute_type == "column"
                        and attribute_def.table.find_source_name()
                        == table.find_source_name()
                        and attribute_def.column_name()
                        not in related_def.get("columns", [])
                    ):
                        c = MeltanoColumn(table=table)
                        c.copy_metadata(attribute_def)
                        table.add_column_with_filter(c)

                # Add the aggregates
                for aggregate in related_def.get("aggregates", []):
                    aggregate_def = table_def.get_aggregate(aggregate)
                    if aggregate_def:
                        a = MeltanoAggregate(table=table)
                        a.copy_metadata(aggregate_def)
                        table.add_aggregate(a)
                    else:
                        raise ParseError(
                            f"Requested aggregate {table.name}.{aggregate} is not defined in the design"
                        )

                # Add the timeframes
                for timeframe in related_def.get("timeframes", []):
                    requested_periods = timeframe.get("periods", None)

                    timeframe_def = table_def.get_timeframe(timeframe.get("name", None))
                    if timeframe_def:
                        t = MeltanoTimeframe(table=table)
                        t.copy_metadata(timeframe_def, requested_periods)
                        table.add_timeframe(t)
                    else:
                        raise ParseError(
                            f"Requested timeframe {table.name}.{timeframe.get('name', None)} is not defined in the design"
                        )

                # Find the primary keys of the Table definition in the Design
                #  that have not been requested in the Query definition
                #  and add them in the Table definition in order to use them
                #  in the SQL generation (we need them in order to select distinct)
                for pk in table_def.primary_keys():
                    if (
                        table.get_column(pk.name) is None
                        and table.get_aggregate_by_column_name(pk.name) is None
                    ):
                        c = MeltanoColumn(table=table)
                        c.copy_metadata(pk)
                        table.optional_pkeys.append(c)

                self.tables.append(table)

                # Add tables that have at least a Column, Aggregate or Timeframe
                #  or a filter over a non selected column
                #  requested in the Query definition
                if related_def is not primary_definition and (
                    table.columns()
                    or table.aggregates()
                    or table.timeframes()
                    or table.unselected_columns_with_filters()
                ):
                    # Find the related joins needed to gather this data
                    #  Using the design graph, we can find the shortest path to a
                    #  table from the primary table up to any table in the design.
                    #
                    #  base_table → (join tables...) → target_table
                    new_joins = self.joins_for_table(
                        design_graph, primary_definition["name"], related_def["name"]
                    )

                    # filter out already defined joins
                    new_joins = list(
                        filter(
                            lambda t: t not in [j["name"] for j in self.join_order],
                            new_joins,
                        )
                    )

                    for join_name in new_joins:
                        if join_name == primary_definition["name"]:
                            continue

                        try:
                            join = self.design.get_join(join_name)
                            join_on = self.design_helper.join_for(join._definition)[
                                "on"
                            ]

                            self.join_order.append(
                                {
                                    "name": join_name,
                                    "table": join.related_table["name"],
                                    "on": join_on,
                                }
                            )
                        except StopIteration:
                            raise ParseError(
                                f"Requested join {join} is not defined in the design"
                            )

                # Take care care for the single table, no join query case
                if not self.join_order:
                    self.join_order.append(
                        {
                            "name": primary_definition["name"],
                            "table": table.name,
                            "on": None,
                        }
                    )
            else:
                raise ParseError(
                    f"Requested table {related_table.get('name', None)} is not defined in the design"
                )

    def joins_for_table(self, design_graph, source_table: str, target_table: str):
        return nx.shortest_path(design_graph, source=source_table, target=target_table)

    def needs_hda(self) -> bool:
        """
        Check if this MeltanoQuery needs to generate an HDA query or not
        """
        has_join_columns = next(
            filter(lambda x: len(x.get("columns", [])), self._definition["joins"]), None
        )

        has_join_timeframes = next(
            filter(lambda x: len(x.get("timeframes", [])), self._definition["joins"]),
            None,
        )

        has_join_aggregates = next(
            filter(lambda x: len(x.get("aggregates", [])), self._definition["joins"]),
            None,
        )

        has_base_table_aggregate = len(self._definition.get("aggregates", []))

        if not has_join_columns and not has_join_timeframes and not has_join_aggregates:
            # A query that references no columns or aggregates from the joined
            #  tables does not need an HDA
            return False
        else:
            # Even if there are joins needed, switch to an HDA only if aggregates
            #  will be computed.
            return has_base_table_aggregate or has_join_aggregates

    def get_query(self) -> Tuple:
        """
        Return the SQL query for this Query definition.

        Returns a Tuple (sql, query_attributes, aggregate_columns)
        - sql: A string with the SQL:1999 compatible query
        - query_attributes: Array of hashes describing the attributes in the
           final result in the same order as the one defined by the query.
           Keys included in the hash:
           {table_name, source_name, attribute_name, attribute_label, attribute_type}
        - aggregate_columns: Array of hashes describing the aggregate columns.
           Keys included in the hash: {id, label, source}
        """
        if self.needs_hda():
            sql, query_attributes, aggregate_columns = self.hda_query()
        else:
            sql, query_attributes, aggregate_columns = self.single_table_query()

        if sql:
            sql = sql + ";"
        else:
            raise EmptyQuery

        return (sql, query_attributes, aggregate_columns)

    def single_table_query(self) -> Tuple:
        """
        Build a simple, no-join SQL query for this Query definition.

        Returns a Tuple (sql, query_attributes, aggregate_columns)
        - sql: A string with the SQL:1999 compatible query
        - query_attributes: Array of hashes describing the attributes in the
           final result in the same order as the one defined by the query.
           Keys included in the hash:
           {table_name, source_name, attribute_name, attribute_label, attribute_type}
        - aggregate_columns: Array of hashes describing the aggregate columns.
           Keys included in the hash: {id, label, source}
        """

        # Lists with the column names and headers of the final result when the
        #  query is executed
        query_attributes = []
        aggregate_columns = []

        select = []
        select_aggregates = []
        group_by_attributes = []

        where = []
        having = []

        base_query = Query

        # Start By building the Join
        for join in self.join_order:
            table = next(
                iter([t for t in self.tables if t.find_source_name() == join["name"]]),
                None,
            )
            # Create a pypika Table based on the Table's name
            pika_table = Table(table.sql_table_name, alias=table.find_source_name())

            if join["on"] is None:
                base_query = base_query.from_(pika_table)
            else:
                base_query = base_query.join(pika_table).on(join["on"])

            # Add all columns in the SELECT clause and as group_by attributes
            for c in table.columns():
                select.append(Field(c.column_name(), table=pika_table, alias=c.alias()))
                group_by_attributes.append(Field(c.alias()))

                query_attributes.append(
                    {
                        "key": c.alias(),
                        "table_name": table.name,
                        "source_name": table.find_source_name(),
                        "attribute_name": c.name,
                        "attribute_label": c.label,
                        "attribute_type": c.type,
                    }
                )

            # Add the WHERE clauses using the column filters for this table
            for c in table.columns() + table.unselected_columns_with_filters():
                matching_criteria = [
                    f.criterion(Field(c.column_name(), table=pika_table))
                    for f in self.column_filters
                    if f.match(c)
                ]
                where.extend(matching_criteria)

            # Add the timeframe columns in the SELECT clause and as group_by attributes
            for tp in table.timeframe_periods():
                select.append(tp.qualified_sql(pika_table=pika_table))
                group_by_attributes.append(Field(tp.alias()))

                query_attributes.append(
                    {
                        "key": tp.alias(),
                        "table_name": table.name,
                        "source_name": table.find_source_name(),
                        "attribute_name": tp.column_name(),
                        "attribute_label": tp.attribute_label(),
                        "attribute_type": tp.type,
                    }
                )

            # Generate the Aggregate Clauses
            for a in table.aggregates():
                select_aggregates.append(a.qualified_sql(pika_table=pika_table))

                aggregate_columns.append(
                    {"id": a.alias(), "label": a.label, "source": table.name}
                )
                query_attributes.append(
                    {
                        "key": a.alias(),
                        "table_name": table.name,
                        "source_name": table.find_source_name(),
                        "attribute_name": a.name,
                        "attribute_label": a.label,
                        "attribute_type": a.type,
                    }
                )

                # Also check if there is a filter for this aggregate and add it
                #  in the HAVING clase
                for f in self.aggregate_filters:
                    if f.match(a):
                        field = a.qualified_sql(pika_table=pika_table)
                        # remove the alias before adding it to the having clause
                        field.alias = None
                        having.append(f.criterion(field))

        # Generate the query
        no_join_query = base_query.select(*select).select(*select_aggregates)

        if where:
            no_join_query = no_join_query.where(Criterion.all(where))

        no_join_query = no_join_query.groupby(*group_by_attributes)

        if having:
            # The check is necessary as PyPika does not accept in having()
            #  the EmptyCriterion that is generated by an empty array
            no_join_query = no_join_query.having(Criterion.all(having))

        # Add the Order By clause(s) for the final query
        if self.order:
            for order_clause in self.order:
                if order_clause.get("direction", None) == "desc":
                    order = Order.desc
                else:
                    order = Order.asc

                key = order_clause.get("key", None)
                if key is not None:
                    # Will raise ParseError if not found
                    query_attribute = self.design.find_attribute(key)
                else:
                    source_name = order_clause.get("source_name")
                    attribute_name = order_clause.get("attribute_name")

                    # Will raise ParseError if not found
                    table_def = self.design.find_table(source_name)
                    query_attribute = table_def.get_attribute(attribute_name)

                # this only works if the field is present in the Select statement
                orderby_field = Field(query_attribute.alias())

                no_join_query = no_join_query.orderby(orderby_field, order=order)
        else:
            # By default order by all the Group By attributes asc
            order = Order.asc
            for field in group_by_attributes:
                no_join_query = no_join_query.orderby(field, order=order)

        # Add a Limit (by default 50)
        no_join_query = no_join_query.limit(self.limit or 50)

        final_query = self.add_schema_to_query(str(no_join_query))

        return (final_query, query_attributes, aggregate_columns)

    def hda_query(self) -> Tuple:
        """
        Build the HDA SQL query for this Query definition.

        For more info on the process followed check the following:
        https://gitlab.com/meltano/meltano/issues/286
        https://gitlab.com/meltano/meltano/issues/344

        Returns a Tuple (sql, query_attributes, aggregate_columns)
        - sql: A string with the hda_query as a SQL:1999 compatible query
        - query_attributes: Array of hashes describing the attributes in the
           final result in the same order as the one defined by the query.
           Keys included in the hash:
           {table_name, source_name, attribute_name, attribute_label, attribute_type}
        - aggregate_columns: Array of hashes describing the aggregate columns.
           Keys included in the hash: {id, label, source}
        """

        # Lists with the column names and headers of the final result when the
        #  HDA query is executed
        query_attributes = []
        aggregate_columns = []

        where = []

        # Build the base_join table
        base_join_query = Query

        # Start By building the Join
        for join in self.join_order:
            table = next(
                iter([t for t in self.tables if t.find_source_name() == join["name"]]),
                None,
            )

            # Create a pypika Table based on the Table's name
            db_table = Table(table.sql_table_name, alias=table.find_source_name())

            if join["on"] is None:
                base_join_query = base_join_query.from_(db_table)
            else:
                base_join_query = base_join_query.join(db_table).on(join["on"])

        # precompute the group_by attributes that will be used in all intermediary
        # base_XXX and XXX_stats with clauses
        group_by_attributes = [c.alias() for t in self.tables for c in t.columns()]

        # Then Select all the columns that will be required in next steps
        # For each table all base columns, columns that will be used in aggregates,
        #  timeframes and the primary keys in case they are not already included
        select = []
        for table in self.tables:
            pika_table = Table(table.sql_table_name, alias=table.find_source_name())

            # Add the WHERE clauses using the column filters for this table
            for c in table.columns() + table.unselected_columns_with_filters():
                matching_criteria = [
                    f.criterion(Field(c.column_name(), table=pika_table))
                    for f in self.column_filters
                    if f.match(c)
                ]
                where.extend(matching_criteria)

            if not (table.columns() or table.aggregates() or table.timeframes()):
                # Skip tables that don't have at least a Column, Aggregate or
                #  Timeframe requested in the Query definition
                continue

            # We have to track wich columns have been added to the SELECT clause
            #  for being used in Aggregates as it is valid to have multiple
            #  Aggregates over the same column (e.g. AVG(price), SUM(price))
            # In that case, we don't want to add the column multiple times in the
            #  select clause cause will cause an Error in followup with clauses:
            #  Error: column reference {COLUMN} is ambiguous
            # There is also the very rare but syntactically correct case of adding
            #  a column both as a group by column and using an Aggregate.
            groupby_columns_selected = set()
            aggregate_columns_selected = set()

            for c in table.columns() + table.optional_pkeys:
                select.append(Field(c.column_name(), table=pika_table, alias=c.alias()))
                groupby_columns_selected.add(c.column_name())

            for a in table.aggregates():
                if (a.column_name() in groupby_columns_selected) or (
                    a.column_name() in aggregate_columns_selected
                ):
                    continue

                select.append(
                    Field(a.column_name(), table=pika_table, alias=a.column_alias())
                )
                aggregate_columns_selected.add(a.column_name())

            for tp in table.timeframe_periods():
                select.append(tp.qualified_sql(pika_table=pika_table))
                group_by_attributes.append(tp.alias())

                query_attributes.append(
                    {
                        "key": tp.alias(),
                        "table_name": table.name,
                        "source_name": table.find_source_name(),
                        "attribute_name": tp.column_name(),
                        "attribute_label": tp.attribute_label(),
                        "attribute_type": tp.type,
                    }
                )

        base_join_query = base_join_query.select(*select)

        if where:
            base_join_query = base_join_query.where(Criterion.all(where))

        # Iterativelly build the HDA query by starting with the base_join
        #  and then adding base_XXX and XXX_stats with clauses for each table
        #  involved in this query
        hda_query = Query.with_(base_join_query, "base_join")
        result_query = Query

        result_query_base_table = None
        for j in self.join_order:
            # Find the table definition in the Query
            table = next(
                (t for t in self.tables if t.find_source_name() == j["name"]), None
            )

            # Skip tables that don't have at least a Column, Aggregate or
            #  Timeframe requested in the Query definition
            if not (table.columns() or table.aggregates() or table.timeframes()):
                continue

            # Generate the base_XXX with clause
            base_db_table = f"base_{j['name']}"

            select = list(
                set(
                    group_by_attributes
                    + [c.alias() for c in table.optional_pkeys]
                    + [c.column_alias() for c in table.aggregates()]
                )
            )

            base_query = Query.from_("base_join").select(*select).distinct()
            hda_query = hda_query.with_(base_query, base_db_table)

            # Generate the XXX_stats with clause
            stats_db_table = f"{j['name']}_stats"

            select = group_by_attributes
            select_aggregates = [
                a.qualified_sql(base_db_table) for a in table.aggregates()
            ]

            stats_query = (
                Query.from_(base_db_table)
                .select(*select)
                .select(*select_aggregates)
                .groupby(*group_by_attributes)
            )

            # Add a Having Clause
            having = []

            for a in table.aggregates():
                for f in self.aggregate_filters:
                    if f.match(a):
                        field = a.qualified_sql(base_db_table)
                        # remove the alias before adding it to the having clause
                        field.alias = None
                        having.append(f.criterion(field))

            if having:
                stats_query = stats_query.having(Criterion.all(having))

            # Add the Stats with clause to the hda query
            hda_query = hda_query.with_(stats_query, stats_db_table)

            stats_pika_table = Table(stats_db_table, alias=stats_db_table)

            if result_query_base_table:
                if group_by_attributes:
                    result_query = result_query.join(stats_pika_table).on_field(
                        *group_by_attributes
                    )
                else:
                    result_query = result_query.join(stats_pika_table).cross()
            else:
                result_query = result_query.from_(stats_pika_table)

                # Add the group_by attributes once in the result query with base
                #  table the first stats table
                result_group_by_attributes = []
                for a in group_by_attributes:
                    result_group_by_attributes.append(
                        Field(a, table=stats_pika_table, alias=a)
                    )

                result_query = result_query.select(*result_group_by_attributes)

                for t in self.tables:
                    for c in t.columns():
                        query_attributes.append(
                            {
                                "key": c.alias(),
                                "table_name": t.name,
                                "source_name": t.find_source_name(),
                                "attribute_name": c.name,
                                "attribute_label": c.label,
                                "attribute_type": c.type,
                            }
                        )

                result_query_base_table = stats_db_table

            # Add the aggregates specific for this table to the result query
            table_aggregates = []
            for a in table.aggregates():
                table_aggregates.append(
                    Field(a.alias(), table=stats_pika_table, alias=a.alias())
                )

                aggregate_columns.append(
                    {"id": a.alias(), "label": a.label, "source": table.name}
                )

                query_attributes.append(
                    {
                        "key": a.alias(),
                        "table_name": table.name,
                        "source_name": table.find_source_name(),
                        "attribute_name": a.name,
                        "attribute_label": a.label,
                        "attribute_type": a.type,
                    }
                )

            result_query = result_query.select(*table_aggregates)

        # Add the result and the Limit, Order By clauses to the final Query
        results_pika_table = Table("result", alias="result")

        hda_query = (
            hda_query.with_(result_query, "result")
            .from_(results_pika_table)
            .select("*")
            .limit(self.limit or 50)
        )

        # Add the Order By clause(s) for the final query
        if self.order:
            for order_clause in self.order:
                if order_clause.get("direction", None) == "desc":
                    order = Order.desc
                else:
                    order = Order.asc

                key = order_clause.get("key", None)
                if key is not None:
                    # Will raise ParseError if not found
                    query_attribute = self.design.find_attribute(key)
                else:
                    source_name = order_clause.get("source_name")
                    attribute_name = order_clause.get("attribute_name")

                    # Will raise ParseError if not found
                    table_def = self.design.find_table(source_name)
                    query_attribute = table_def.get_attribute(attribute_name)

                # this only works if the field is present in the Select statement
                orderby_field = Field(query_attribute.alias(), table=results_pika_table)

                hda_query = hda_query.orderby(orderby_field, order=order)
        else:
            # By default order by all the Group By attributes asc
            order = Order.asc
            for attr in group_by_attributes:
                field = Field(attr, table=results_pika_table)
                hda_query = hda_query.orderby(field, order=order)

        final_query = self.add_schema_to_query(str(hda_query))

        return (final_query, query_attributes, aggregate_columns)

    def add_schema_to_query(self, query: str) -> str:
        """
        This is a temporary solution to address an issue in the way pypika
        validates join conditions.

        if you have a schema and an alias, the following SQL is perfectly fine:
            FROM "analytics"."table1" "table1"
            JOIN "analytics"."table2" "table2"
              ON "table1"."id1"="table2"."id2"

        But unfortunately, pypika considers that as not valid:
            pypika.utils.JoinException: Invalid join criterion.
            One field is required from the joined item and another from the
                selected table or an existing join.
            Found ["table1" "table1", "table2" "table2"]

        Until they allow for "join .. on" clauses to be able to use the alias
          of the tables, we have to construct the query without a schema and
          then add the schema by replacing the table name with schema.table
          Pretty ugly but the best we got at the moment.
        """
        if self.schema is None or self.schema == "":
            return query

        schema_query = query
        for join in self.join_order:
            table = next(
                (t for t in self.tables if t.find_source_name() == join["name"]), None
            )

            if join["on"] is None:
                schema_query = schema_query.replace(
                    f'FROM "{table.sql_table_name}"',
                    f'FROM "{self.schema}"."{table.sql_table_name}"',
                )
            else:
                schema_query = schema_query.replace(
                    f'JOIN "{table.sql_table_name}"',
                    f'JOIN "{self.schema}"."{table.sql_table_name}"',
                )

        return schema_query



"""SingerTarget and supporting classes.

This module contains the SingerTarget class as well as a supporting BookmarkWriter class.
"""
import json
import logging
from datetime import datetime
from typing import List

from meltano.core.behavior.hookable import hook
from meltano.core.db import project_engine
from meltano.core.job import Job, Payload
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.setting_definition import SettingDefinition

from . import PluginType, SingerPlugin

logger = logging.getLogger(__name__)


class BookmarkWriter:
    """A basic bookmark writer suitable for use as an output handler."""

    def __init__(self, job: Job, session: object, payload_flag: int = Payload.STATE):
        """Bookmark writer with a writelines implementation to support ingesting and persisting state messages.

        Args:
            job: meltano elt job associated with this invocation and who's state will be updated.
            session: SQLAlchemy session/engine object to be used to update state.
            payload_flag: a valid payload flag, one of Payload.STATE or Payload.INCOMPLETE_STATE.
        """
        self.job = job
        self.session = session
        self.payload_flag = payload_flag

    def writeline(self, line: str):
        """Persist a state entry.

        Args:
            line: raw json state line to decode/store
        """
        if self.job is None:
            logging.info(
                "Running outside a Job context: incremental state could not be updated."
            )
            return

        new_state = {}
        try:
            new_state = json.loads(line)
        except Exception:
            logging.warning(
                "Received state is invalid, incremental state has not been updated"
            )

        job = self.job
        job.payload["singer_state"] = new_state
        job.payload_flags |= self.payload_flag
        try:
            job.save(self.session)
        except Exception:
            logging.warning(
                "Unable to persist state, or received state is invalid, incremental state has not been updated"
            )
        else:
            logging.info(f"Incremental state has been updated at {datetime.utcnow()}.")
            logging.debug(f"Incremental state: {new_state}")


class SingerTarget(SingerPlugin):
    __plugin_type__ = PluginType.LOADERS

    EXTRA_SETTINGS = [
        SettingDefinition(name="_dialect", value="$MELTANO_LOADER_NAMESPACE"),
        SettingDefinition(name="_target_schema", value="$MELTANO_LOAD_SCHEMA"),
    ]

    def exec_args(self, plugin_invoker):
        args = ["--config", plugin_invoker.files["config"]]

        return args

    @property
    def config_files(self):
        return {"config": f"target.{self.instance_uuid}.config.json"}

    @property
    def output_files(self):
        return {"state": "new_state.json"}

    @hook("before_invoke")
    def setup_bookmark_writer_hook(
        self, plugin_invoker: PluginInvoker, exec_args: List[str]
    ):
        """Before invoke hook to trigger setting up the bookmark writer for this target.

        Args:
            plugin_invoker: The invocation handler of the plugin instance.
            exec_args: List of subcommand/args that we where invoked with.
        """
        if exec_args is None:
            exec_args = []

        if "--discover" in exec_args or "--help" in exec_args:
            return

        self.setup_bookmark_writer(plugin_invoker)

    def setup_bookmark_writer(self, plugin_invoker: PluginInvoker):
        """Configure the bookmark writer as an additional output handler on the invoker.

        This leverages calling back to PluginInvokers.add_output_handler to attach an additional
        output handler (the BookmarkWriter) to handle persisting state messages.

        Args:
            plugin_invoker: The invocation handler who's add_out_handler method will be called to attach the bookmark writer
            as an additional output handler.
        """
        elt_context = plugin_invoker.context
        incomplete_state = elt_context.full_refresh and elt_context.select_filter
        payload_flag = Payload.INCOMPLETE_STATE if incomplete_state else Payload.STATE

        plugin_invoker.add_output_handler(
            plugin_invoker.StdioSource.STDOUT,
            BookmarkWriter(elt_context.job, elt_context.session, payload_flag),
        )



import json
from datetime import date, datetime
from enum import Enum
from typing import List

from .behavior import NameEq
from .behavior.canonical import Canonical
from .error import Error
from .utils import flatten, nest_object, to_env_var, truthy, uniques_in

VALUE_PROCESSORS = {
    "nest_object": nest_object,
    "upcase_string": lambda s: s.upper(),
    "stringify": lambda s: s if isinstance(s, str) else json.dumps(s),
}


class EnvVar:
    def __init__(self, definition):
        key = definition
        negated = False

        if definition.startswith("!"):
            key = definition[1:]
            negated = True

        self.key = key
        self.negated = negated

    @property
    def definition(self):
        prefix = "!" if self.negated else ""
        return f"{prefix}{self.key}"

    def get(self, env):
        if self.negated:
            return str(not truthy(env[self.key]))
        else:
            return env[self.key]


class SettingMissingError(Error):
    """Occurs when a setting is missing."""

    def __init__(self, name: str):
        super().__init__(f"Cannot find setting {name}")


class YAMLEnum(str, Enum):
    """Serializable Enum class."""

    def __str__(self):
        """Return as string."""
        return self.value

    @staticmethod
    def yaml_representer(dumper, obj):
        """Represent as yaml."""
        return dumper.represent_scalar("tag:yaml.org,2002:str", str(obj))


class SettingKind(YAMLEnum):
    """Supported setting kinds."""

    STRING = "string"
    INTEGER = "integer"
    BOOLEAN = "boolean"
    DATE_ISO8601 = "date_iso8601"
    EMAIL = "email"
    PASSWORD = "password"  # noqa: S105
    OAUTH = "oauth"
    OPTIONS = "options"
    FILE = "file"
    ARRAY = "array"
    OBJECT = "object"
    HIDDEN = "hidden"


class SettingDefinition(NameEq, Canonical):
    def __init__(
        self,
        name: str = None,
        aliases: List[str] = [],
        env: str = None,
        env_aliases: List[str] = [],
        kind: SettingKind = None,
        value=None,
        label: str = None,
        documentation: str = None,
        description: str = None,
        tooltip: str = None,
        options: list = [],
        oauth: dict = {},
        placeholder: str = None,
        protected: bool = None,
        env_specific: bool = None,
        custom: bool = False,
        value_processor=None,
        value_post_processor=None,
        **attrs,
    ):
        super().__init__(
            # Attributes will be listed in meltano.yml in this order:
            name=name,
            aliases=aliases,
            env=env,
            env_aliases=env_aliases,
            kind=SettingKind(kind) if kind else None,
            value=value,
            label=label,
            documentation=documentation,
            description=description,
            tooltip=tooltip,
            options=options,
            oauth=oauth,
            placeholder=placeholder,
            protected=protected,
            env_specific=env_specific,
            value_processor=value_processor,
            value_post_processor=value_post_processor,
            _custom=custom,
            **attrs,
        )

        self._verbatim.add("value")

    @classmethod
    def from_missing(cls, defs, config, **kwargs):
        flat_config = flatten(config, "dot")

        names = {s.name for s in defs}

        # Create custom setting definitions for unknown keys
        return [
            SettingDefinition.from_key_value(k, v, **kwargs)
            for k, v in flat_config.items()
            if k not in names
        ]

    @classmethod
    def from_key_value(cls, key, value, custom=True, default=False):
        kind = None
        if isinstance(value, bool):
            kind = SettingKind.BOOLEAN
        elif isinstance(value, int):
            kind = SettingKind.INTEGER
        elif isinstance(value, dict):
            kind = SettingKind.OBJECT
        elif isinstance(value, list):
            kind = SettingKind.ARRAY

        attrs = {
            "name": key,
            "kind": kind,
            "custom": custom,
            "value": value if default else None,
        }

        return cls(**attrs)

    @property
    def is_extra(self):
        return self.name.startswith("_")

    @property
    def is_custom(self):
        """Return whether the setting is custom, i.e. user-defined in `meltano.yml`."""
        return self._custom

    @property
    def is_redacted(self):
        return self.kind in {SettingKind.PASSWORD, SettingKind.OAUTH}

    def env_vars(self, prefixes: [str], include_custom=True):
        """Return environment variables with the provided prefixes."""
        env_keys = []

        if self.env and include_custom:
            env_keys.append(self.env)

        env_keys.extend(to_env_var(prefix, self.name) for prefix in prefixes)

        if include_custom:
            env_keys.extend(alias for alias in self.env_aliases)

        return [EnvVar(key) for key in uniques_in(env_keys)]

    def cast_value(self, value):
        if isinstance(value, date) or isinstance(value, datetime):
            value = value.isoformat()

        if isinstance(value, str):
            if self.kind == SettingKind.BOOLEAN:
                return truthy(value)
            elif self.kind == SettingKind.INTEGER:
                return int(value)
            elif self.kind == SettingKind.OBJECT:
                value = json.loads(value)
                if not isinstance(value, dict):
                    raise ValueError(f"JSON value '{value}' is not an object")
            elif self.kind == SettingKind.ARRAY:
                value = json.loads(value)
                if not isinstance(value, list):
                    raise ValueError(f"JSON value '{value}' is not an array")

        processor = self.value_processor
        if value is not None and processor:
            if isinstance(processor, str):
                processor = VALUE_PROCESSORS[processor]
            value = processor(value)

        return value

    def post_process_value(self, value):
        processor = self.value_post_processor
        if value is not None and processor:
            if isinstance(processor, str):
                processor = VALUE_PROCESSORS[processor]
            value = processor(value)

        return value

    def stringify_value(self, value):
        if isinstance(value, str):
            return value

        if not self.kind or self.kind == SettingKind.STRING:
            return str(value)

        return json.dumps(value)



from typing import Dict

from meltano.core.plugin import PluginType

from .base import *
from .catalog import ListExecutor, SelectExecutor
from .tap import SingerTap
from .target import BookmarkWriter, SingerTarget



from typing import Dict, Iterable, List

from meltano.core.plugin import BasePlugin
from meltano.core.plugin.project_plugin import ProjectPlugin
from meltano.core.plugin_discovery_service import PluginDiscoveryService
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService
from meltano.core.project_settings_service import ProjectSettingsService
from meltano.core.setting_definition import SettingDefinition
from meltano.core.settings_service import (
    REDACTED_VALUE,
    SettingMissingError,
    SettingsService,
    SettingValueStore,
)


class PluginSettingsService(SettingsService):
    def __init__(
        self,
        project: Project,
        plugin: ProjectPlugin,
        *args,
        plugins_service: ProjectPluginsService = None,
        **kwargs,
    ):
        super().__init__(project, *args, **kwargs)

        self.plugin = plugin
        self.plugins_service = plugins_service or ProjectPluginsService(self.project)

        project_settings_service = ProjectSettingsService(
            self.project, config_service=self.plugins_service.config_service
        )

        self.env_override = {
            **project_settings_service.env,
            **project_settings_service.as_env(),
            **self.env_override,
            **self.plugin.info_env,
        }

        self._inherited_settings_service = None

    @property
    def label(self):
        return f"{self.plugin.type.descriptor} '{self.plugin.name}'"

    @property
    def docs_url(self):
        return self.plugin.docs

    def setting_env_vars(self, setting_def, for_writing=False):
        return setting_def.env_vars(
            self.plugin.env_prefixes(for_writing=for_writing),
            include_custom=self.plugin.is_shadowing or for_writing,
        )

    @property
    def db_namespace(self):
        """Return namespace for setting value records in system database."""
        # "default" is included for legacy reasons
        return ".".join((self.plugin.type, self.plugin.name, "default"))

    @property
    def setting_definitions(self):
        """Return definitions of supported settings."""
        return self.plugin.settings_with_extras

    @property
    def meltano_yml_config(self):
        """Return current configuration in `meltano.yml`."""
        return self.plugin.config_with_extras

    def update_meltano_yml_config(self, config_with_extras):
        """Update configuration in `meltano.yml`."""
        self.plugin.config_with_extras = config_with_extras
        self.plugins_service.update_plugin(self.plugin)

    @property
    def inherited_settings_service(self):
        """Return settings service to inherit configuration from."""
        parent_plugin = self.plugin.parent
        if not isinstance(parent_plugin, ProjectPlugin):
            return None

        if self._inherited_settings_service is None:
            self._inherited_settings_service = self.__class__(
                self.project,
                parent_plugin,
                env_override=self.env_override,
                plugins_service=self.plugins_service,
            )
        return self._inherited_settings_service

    def process_config(self, config):
        """Process configuration dictionary to be passed to plugin."""
        return self.plugin.process_config(config)



from collections import namedtuple
from typing import Optional

from meltano.core.job import Job
from meltano.core.plugin import PluginRef, PluginType
from meltano.core.plugin.settings_service import PluginSettingsService
from meltano.core.plugin_invoker import invoker_factory
from meltano.core.project import Project
from meltano.core.project_plugins_service import ProjectPluginsService


class PluginContext(namedtuple("PluginContext", "plugin settings_service session")):
    def __getattr__(self, attr):
        return getattr(self.plugin, attr)

    def get_config(self, name, **kwargs):
        return self.settings_service.get(name, session=self.session, **kwargs)

    def config_dict(self, **kwargs):
        return self.settings_service.as_dict(session=self.session, **kwargs)

    def config_env(self, **kwargs):
        return self.settings_service.as_env(session=self.session, **kwargs)

    @property
    def env(self):
        return {**self.plugin.info_env, **self.config_env()}


class ELTContext:
    def __init__(
        self,
        project,
        job: Optional[Job] = None,
        session=None,
        extractor: Optional[PluginContext] = None,
        loader: Optional[PluginContext] = None,
        transform: Optional[PluginContext] = None,
        transformer: Optional[PluginContext] = None,
        only_transform: Optional[bool] = False,
        dry_run: Optional[bool] = False,
        full_refresh: Optional[bool] = False,
        select_filter: Optional[list] = [],
        catalog: Optional[str] = None,
        state: Optional[str] = None,
        plugins_service: ProjectPluginsService = None,
    ):
        self.project = project
        self.job = job
        self.session = session

        self.extractor = extractor
        self.loader = loader
        self.transform = transform
        self.transformer = transformer

        self.only_transform = only_transform
        self.dry_run = dry_run
        self.full_refresh = full_refresh
        self.select_filter = select_filter
        self.catalog = catalog
        self.state = state

        self.plugins_service = plugins_service or ProjectPluginsService(project)

    @property
    def elt_run_dir(self):
        if self.job:
            return self.project.job_dir(self.job.job_id, str(self.job.run_id))

    def invoker_for(self, plugin_type):
        plugin_contexts = {
            PluginType.EXTRACTORS: self.extractor,
            PluginType.LOADERS: self.loader,
            PluginType.TRANSFORMERS: self.transformer,
        }

        plugin_context = plugin_contexts[plugin_type]

        return invoker_factory(
            self.project,
            plugin_context.plugin,
            context=self,
            run_dir=self.elt_run_dir,
            plugins_service=self.plugins_service,
            plugin_settings_service=plugin_context.settings_service,
        )

    def extractor_invoker(self):
        return self.invoker_for(PluginType.EXTRACTORS)

    def loader_invoker(self):
        return self.invoker_for(PluginType.LOADERS)

    def transformer_invoker(self):
        return self.invoker_for(PluginType.TRANSFORMERS)


class ELTContextBuilder:
    def __init__(self, project: Project, plugins_service: ProjectPluginsService = None):
        self.project = project
        self.plugins_service = plugins_service or ProjectPluginsService(project)

        self._session = None
        self._job = None

        self._extractor = None
        self._loader = None
        self._transform = None
        self._transformer = None

        self._only_transform = False
        self._dry_run = False
        self._full_refresh = False
        self._select_filter = None
        self._catalog = None
        self._state = None

    def with_session(self, session):
        self._session = session

        return self

    def with_job(self, job: Job):
        self._job = job

        return self

    def with_extractor(self, extractor_name: str):
        self._extractor = PluginRef(PluginType.EXTRACTORS, extractor_name)

        return self

    def with_loader(self, loader_name: str):
        self._loader = PluginRef(PluginType.LOADERS, loader_name)

        return self

    def with_transform(self, transform: str):
        if transform == "skip":
            return self

        if transform not in ("run", "only"):
            self._transform = PluginRef(PluginType.TRANSFORMS, transform)

        return self.with_transformer("dbt")

    def with_transformer(self, transformer_name: str):
        self._transformer = PluginRef(PluginType.TRANSFORMERS, transformer_name)

        return self

    def with_only_transform(self, only_transform):
        self._only_transform = only_transform

        return self

    def with_dry_run(self, dry_run):
        self._dry_run = dry_run

        return self

    def with_full_refresh(self, full_refresh):
        self._full_refresh = full_refresh

        return self

    def with_select_filter(self, select_filter):
        self._select_filter = select_filter

        return self

    def with_catalog(self, catalog):
        self._catalog = catalog

        return self

    def with_state(self, state):
        self._state = state

        return self

    def plugin_context(self, plugin_ref: PluginRef, env={}, config={}):
        plugin = self.plugins_service.get_plugin(plugin_ref)

        return PluginContext(
            plugin=plugin,
            settings_service=PluginSettingsService(
                self.project,
                plugin,
                plugins_service=self.plugins_service,
                env_override=env,
                config_override=config,
            ),
            session=self._session,
        )

    def context(self) -> ELTContext:
        env = {}

        extractor = None
        if self._extractor:
            config = {}
            if self._select_filter:
                config["_select_filter"] = self._select_filter
            if self._catalog:
                config["_catalog"] = self._catalog
            if self._state:
                config["_state"] = self._state

            extractor = self.plugin_context(self._extractor, config=config)

            env.update(extractor.env)

        loader = None
        if self._loader:
            loader = self.plugin_context(self._loader, env=env.copy())

            env.update(loader.env)

        transform = None
        if self._transform:
            transform = self.plugin_context(self._transform, env=env.copy())

            env.update(transform.env)

        transformer = None
        if self._transformer:
            transformer = self.plugin_context(self._transformer, env=env.copy())

        return ELTContext(
            self.project,
            job=self._job,
            session=self._session,
            extractor=extractor,
            loader=loader,
            transform=transform,
            transformer=transformer,
            only_transform=self._only_transform,
            dry_run=self._dry_run,
            full_refresh=self._full_refresh,
            select_filter=self._select_filter,
            catalog=self._catalog,
            state=self._state,
            plugins_service=self.plugins_service,
        )



from pathlib import Path

import meltano.core.bundle as bundle
import yaml

from .file import FilePlugin


class MeltanoFilePlugin(FilePlugin):
    def __init__(self, discovery=False):
        super().__init__(None, None)
        self._discovery = discovery

    def file_contents(self, project):
        initialize_file = bundle.find("initialize.yml")
        file_contents = {
            Path(relative_path): content
            for relative_path, content in yaml.safe_load(initialize_file.open()).items()
        }
        if self._discovery:
            file_contents["discovery.yml"] = bundle.find("discovery.yml").read_text()
        return file_contents

    def update_config(self, project):
        return {}



import json
import logging
import os
from uuid import uuid4

from meltano.core.behavior.hookable import hook
from meltano.core.db import project_engine
from meltano.core.plugin import BasePlugin
from meltano.core.project import Project
from meltano.core.utils import nest_object


class SingerPlugin(BasePlugin):
    def __init__(self, *args, **kwargs):
        """Canonical class leads to  an error if the UUID is defined here directly. Also, This data attribute must be defined or we'll get errors from Canonical."""
        super().__init__(*args, **kwargs)
        self._instance_uuid: str = None

    def process_config(self, flat_config):
        non_null_config = {k: v for k, v in flat_config.items() if v is not None}
        processed_config = nest_object(non_null_config)
        # Result at this point will contain duplicate entries for nested config
        # options. We need to pop those redundant entries recursively.

        def _pop_non_leaf_keys(nested_config: dict) -> None:  # noqa: WPS430
            """Recursively pop dictionary entries with '.' in their keys."""
            for key, val in list(nested_config.items()):
                if "." in key:
                    nested_config.pop(key)
                elif isinstance(val, dict):  # noqa: WPS220
                    _pop_non_leaf_keys(val)

        _pop_non_leaf_keys(processed_config)
        return processed_config

    @hook("before_configure")
    def before_configure(self, invoker, session):
        config_path = invoker.files["config"]
        with open(config_path, "w") as config_file:
            config = invoker.plugin_config_processed
            json.dump(config, config_file, indent=2)

        logging.debug(f"Created configuration at {config_path}")

    @hook("before_cleanup")
    def before_cleanup(self, invoker):
        config_path = invoker.files["config"]
        config_path.unlink()
        logging.debug(f"Deleted configuration at {config_path}")

    @property
    def instance_uuid(self):
        """Multiple processes running at the same time have a unique value to use."""
        if not self._instance_uuid:
            self._instance_uuid = str(uuid4())
        return self._instance_uuid



import re
from collections import namedtuple
from enum import Enum

from pypika import Case, Field


class SubstitutionType(Enum):
    Unknown = "UNKNOWN"
    Table = "TABLE"
    Column = "COLUMN"
    TableColumn = "TABLE_COLUMN"
    SQLTableName = "SQL_TABLE_NAME"


class Substitution:
    def __init__(self, _input, table, column=None, alias=None):
        self.input = _input
        self.alias = alias
        self.sql = None
        self.table = table

        if not column:
            self.type = "string"
        else:
            self.type = column["type"]

        self.substitution_type = SubstitutionType.Unknown
        self.get_substitution_type()
        self.placeholders = Substitution.placeholder_match(self.input)
        self.set_sql()

    def get_substitution_type(self):
        # trying guess the substitution_type in a cheap way
        if re.search(r"\{\{table\}\}", self.input, re.IGNORECASE):
            self.substitution_type = SubstitutionType.Table
        elif "." in self.input:
            if "SQL_TABLE_NAME" in self.input:
                self.substitution_type = SubstitutionType.SQLTableName
            else:
                self.substitution_type = SubstitutionType.TableColumn
        elif " " not in self.input:
            self.substitution_type = SubstitutionType.Column
        else:
            self.substitution_type = SubstitutionType.Unknown

    @staticmethod
    def placeholder_match(input_):
        outer_pattern = r"(\{\{[\w\.]*\}\})"
        inner_pattern = r"\{\{([\w\.]*)\}\}"
        outer_results = re.findall(outer_pattern, input_)
        inner_results = re.findall(inner_pattern, input_)
        Results = namedtuple("Results", "inner outer")
        return Results(inner=inner_results, outer=outer_results)

    def set_sql(self):
        if self.substitution_type is SubstitutionType.Table:
            self.set_sql_table_type()
        else:
            raise Exception(
                f"Substitution Type {self.substitution_type.value} not implemented yet"
            )

    def set_sql_table_type(self):
        self.sql = self.input.replace(
            self.placeholders.outer[0], self.table._table_name
        )
        (table, field) = self.sql.split(".")

        if not self.alias:
            self.alias = self.sql
        if self.type == "yesno":
            field = Field(field, table=self.table)
            self.sql = Case(alias=self.sql).when(field, "yes").else_("no")
        else:
            self.sql = Field(field, table=self.table, alias=self.sql)



import logging

import click
import psycopg2
import psycopg2.sql
from meltano.core.db import DB, project_engine
from meltano.core.project import Project

from . import cli
from .params import pass_project


@cli.group()
def schema():
    pass


@schema.command()
@click.argument("schema")
@click.argument("roles", nargs=-1, required=True)
@pass_project()
def create(project, schema, roles):
    engine, _ = project_engine(project)

    DB.ensure_schema_exists(engine, schema, grant_roles=roles)



import asyncio
import logging
import os
import sys
from io import StringIO

from meltano.core.db import project_engine
from meltano.core.elt_context import ELTContext
from meltano.core.error import SubprocessError
from meltano.core.logging import capture_subprocess_output
from meltano.core.plugin import PluginType
from meltano.core.plugin_invoker import PluginInvoker
from meltano.core.project import Project

from . import Runner, RunnerError


class DbtRunner(Runner):
    def __init__(self, elt_context: ELTContext):
        self.context = elt_context

    @property
    def project(self):
        return self.context.project

    @property
    def plugin_context(self):
        return self.context.transformer

    async def invoke(self, dbt: PluginInvoker, *args, log=None, **kwargs):
        """Call the dbt executable with the given arguments in a new process."""
        log = log or sys.stderr

        try:
            handle = await dbt.invoke_async(
                *args,
                **kwargs,
                stdout=asyncio.subprocess.PIPE,
                stderr=asyncio.subprocess.PIPE,
            )
        except Exception as err:
            raise RunnerError(f"Cannot start dbt: {err}") from err

        await asyncio.wait(
            [
                capture_subprocess_output(handle.stdout, log),
                capture_subprocess_output(handle.stderr, log),
                handle.wait(),
            ],
            return_when=asyncio.ALL_COMPLETED,
        )

        exitcode = handle.returncode
        if exitcode:
            command = kwargs["command"] or args[0]
            raise RunnerError(
                f"`dbt {command}` failed", {PluginType.TRANSFORMERS: exitcode}
            )

    async def run(self, log=None):
        dbt = self.context.transformer_invoker()

        with dbt.prepared(self.context.session):
            await self.invoke(dbt, log=log, command="clean")
            await self.invoke(dbt, log=log, command="deps")

            cmd = "compile" if self.context.dry_run else "run"
            await self.invoke(dbt, log=log, command=cmd)
